#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Derives a periodic task-set from acyclic CSDF graph model'

import sys
import math
import os.path
import argparse

from CSDFParser import CSDFParser
from ActorModel import ActorModel
from ChannelModel import ChannelModel
import Utilities


verbose = False


class ACSDFModel:
    """ACSDFModel class 

    Represents an acyclic Cyclo-Static DataFlow (CSDF) graph.
    CSDF is originally defined in the following article:
    G. Bilsen, M. Engels, R. Lauwereins, and J. Peperstraete, "Cycle-static dataflow," IEEE Transactions on Signal Processing, vol. 44, no. 2, pp. 397-408, February 1996.
    """

    # The supported start time and buffer computation methods.
    # SUF: Implicit-deadline with sufficient start time and buffer
    # MIN: Constrained-deadline with minimum start time and buffer
    SUF, MIN = range(2)

    def __init__(self):
        """The Constructor"""
        self.actors = []
        self.channels = []
        self.latency = 0
        self.alpha = 0
        self.num_of_levels = 0
        self.lcm = 0
        self.max_q_C = 0
        self.factor = 1
        self.max_token_size = 0
        self.major_cycle = 0
        self.minor_cycle = 0

    def get_max_token_size(self):
        return self.max_token_size

    def set_max_token_size(self, token_size):
        self.max_token_size = token_size

    def add_actor(self, actor_id, name, function, wcet):
        """Adds an actor to the graph"""
        am = ActorModel(self, actor_id, name, function, wcet)
        self.actors.append(am)

    def add_channel(self, channel_id, name, source, destination, production_sequence, consumption_sequence):
        """Adds a channel to the graph"""
        cm = ChannelModel(channel_id, name, source, destination, production_sequence, consumption_sequence)
        self.channels.append(cm)
        for i in self.actors:
            if i.get_actor_id() == source:
                i.add_outport(channel_id)
            if i.get_actor_id() == destination:
                i.add_inport(channel_id)

    def get_actor(self, actor_id):
        """Returns the actor object with the given actor_id. Returns None if no corresponding actor is found"""
        for i in self.actors:
            if (i.get_actor_id() == actor_id):
                return i
        return None

    def get_channel(self, channel_id):
        """Returns the channel object with given channel_id. Returns None if no corresponding channel is found"""
        for i in self.channels:
            if (i.get_channel_id() == channel_id):
                return i
        return None

    def get_actors(self):
        """Returns the set of actor objects"""
        return self.actors

    def get_sink_actors(self):
        levels_actors = self.find_levels()
        return levels_actors[-1]

    def get_channels(self):
        """Returns the set of channel objects"""
        return self.channels

    def get_num_of_actors(self):
        """Returns the number of actors in the graph"""
        return len(self.actors)

    def get_num_of_channels(self):
        """"Returns the number of channels in the graph"""
        return len(self.channels)

    def get_graph_name(self):
        """Returns the graph name"""
        return self.name

    def set_graph_name(self, name):
        """Set the graph name"""
        self.name = name

    def get_factor(self):
        return self.factor

    def set_factor(self, factor):
        assert type(factor) == int
        self.factor = factor

    def get_latency(self):
        """Returns the graph latency"""
        return self.latency

    def update_parameters(self, interconnect_type, word_read_time, word_write_time):
        for a in self.get_actors():
            a.update_execution_time(interconnect_type, word_read_time, word_write_time)
        self.find_minimum_period_vector(self.get_factor())

    def find_minor_cycle(self, lb, major_cycle):
        minor_cycle = lb
        while minor_cycle <= major_cycle:
            if major_cycle % minor_cycle == 0:
                found = True
                for a in self.get_actors():
                    val = minor_cycle + (minor_cycle - Utilities.gcd(minor_cycle, a.get_period()))
                    if val > a.get_deadline():
                        found = False
                        break
                if found:
                    return minor_cycle
            minor_cycle += 1
        return 0

    def compute_cyclic_executive_parameters(self):
        major_cycle = Utilities.lcmv(self.get_period_vector())
        lb = max(self.get_execution_vector())
        minor_cycle = self.find_minor_cycle(lb, major_cycle)
        if minor_cycle != 0:
            assert major_cycle % minor_cycle == 0
            for a in self.get_actors():
                val = minor_cycle + (minor_cycle - Utilities.gcd(minor_cycle, a.get_period()))
                assert val <= a.get_deadline()
            self.major_cycle = major_cycle
            self.minor_cycle = minor_cycle
        else:
            # Unable to find a minor cycle with no preemption
            self.major_cycle = major_cycle
            self.minor_cycle = minor_cycle

    def find_levels(self):
        """Finds the levels of a graph"""
        levels = list()
        actors = self.copy_actors()

        # TODO: Find a better way to check emptiness
        while len(actors) > 0:
            working_set = set()
            for i in actors:
                if len(i.get_inports()) == 0:
                    working_set.add(i)
                    actors.remove(i)
            levels.append(working_set)

            for j in working_set:
                for l in j.get_outports():
                    for k in actors:
                        # TODO: Move the exception handling to the remove methods
                        try:
                            k.get_inports().index(l)
                            k.remove_inport(l)
                        except:
                            pass
        return levels

    def find_minimum_period_vector(self, factor):
        """Finds the minimum period vector and then multiplies it with factor"""
        max_q_C = 0
        for i in self.actors:
            val = i.get_wcet()*i.get_repetition()
            if val > max_q_C:
                max_q_C = val
        self.max_q_C = max_q_C
        self.lcm = Utilities.lcmv(self.get_repetition_vector())
        self.alpha = int(self.lcm * int(math.ceil(float(self.max_q_C)/float(self.lcm))) * factor)
        for i in self.actors:
            period = int(self.alpha / i.get_repetition())
            i.set_period(period)
            i.set_deadline(period) # Initial value of deadline
            assert i.get_period() * i.get_repetition() == self.alpha

        if self.max_q_C % self.lcm == 0:
            self.is_matched_io_rates = True
        else:
            self.is_matched_io_rates = False

    def get_alpha(self):
        """Returns the actor iteration period"""
        return self.alpha

    def get_matched_io_type(self):
        return self.is_matched_io_rates

    def find_earliest_start_time(self, actor, level):
        """Finds the earliest start time of an actor. Invoked when self.method = MIN"""

        assert actor.get_period() * actor.get_repetition() == self.alpha
        alpha = self.alpha
        v_m = None
        if level == 1:
            start_time = 0
            actor.set_start_time(start_time)
        else:
            start_time = 0
            for inport in actor.get_inports():
                channel = self.get_channel(inport)
                assert channel != None
                predecessor = self.get_actor(channel.get_source())
                assert predecessor != None
                candidates = []
                # Use binary search
                t_i = predecessor.get_start_time()
                t_min = 0
                t_max = t_i + alpha
                candidates.append(t_max) # fail-safe value
                if 0 < t_i and actor.cns(channel, 0, t_i) > 0:
                    t_min = t_i
                while t_min < t_max:
                    found = True
                    t_j = (t_min + t_max)//2 # t_j is the pivot value
                    t_prime = max(t_i, t_j)
                    for t_hat in range(0, alpha + 1, actor.get_period()):
                        prd_i = predecessor.prd(channel, t_i , t_prime + t_hat)
                        cns_j = actor.cns(channel, t_j, t_prime + t_hat )
                        net_tokens = prd_i - cns_j
                        if net_tokens < 0:
                            found = False
                            break
                    if not found:
                        t_min = t_j + 1
                    else:
                        candidates.append(t_j)
                        t_max = t_j
                st = min(candidates)
                if st > start_time:
                    start_time = st
                    v_m = predecessor
            if v_m != None:
                actor.set_start_time(start_time)
        return v_m

    def find_earliest_start_time_vector(self, deadline_value):

        assert deadline_value >= 0.0 and deadline_value <= 1.0
        levels = self.find_levels()
        level = 1
        for s in levels:
            for a in s:
                actor = self.get_actor(a.get_actor_id())
                v_m = None
                while True:
                    v_m = self.find_earliest_start_time(actor, level)
                    if v_m != None:
                        v_m.set_deadline(v_m.get_wcet() + int(deadline_value*(v_m.get_period() - v_m.get_wcet())))
                        v_k = self.find_earliest_start_time(actor, level)
                        assert v_k != None
                        if v_k.get_actor_id() != v_m.get_actor_id():
                            pass
                        else:
                            break
                    else:
                        break
            level = level + 1
        for a in self.get_actors():
            if a.get_outports() == []:
                a.set_deadline(a.get_wcet() + int(deadline_value*(a.get_period() - a.get_wcet())))
        self.compute_latency()

    def get_paths(self,input_actor):
        assert input_actor.get_inports() == []
        done = False
        paths = [[] for i in range(len(input_actor.get_outports()))]
        actors = []
        channels = []
        actors.append(input_actor)
        while not done:
            cactors = list(actors)
            for a in cactors:
                for c in a.get_outports():
                    ch = self.get_channel(c)
                    channels.append(ch)
                    actors.append(self.get_actor(ch.get_destination()))
                    for p in paths:
                        if p == [] or p[-1][2] == ch.get_source():
                            p.append((ch.get_channel_id(), ch.get_source(), ch.get_destination()))
                if a.get_outports() != []:
                    actors.remove(a)
            done = True
            for ac in actors:
                if ac.get_outports() != []:
                    done = False
                    break
        return paths

    def compute_latency(self):
        self.latency = 0
        for a in self.get_actors():
            if a.get_outports() == []:
                for b in self.get_actors():
                    if b.get_inports() == []:
                        # Check if there is a path w_{a->b}
                        paths = self.get_paths(b)
                        for path in paths:
                            if path[-1][2] == a.get_actor_id():
                                # Bingo! Output path
                                first_ch = self.get_channel(path[0][0])
                                last_ch = self.get_channel(path[-1][0])
                                g_i = ChannelModel.empty_firings(first_ch.get_production_sequence())
                                f_j = ChannelModel.empty_firings(last_ch.get_consumption_sequence())
                                input_actor = self.get_actor(path[0][1])
                                output_actor = self.get_actor(path[-1][2])
                                latency = output_actor.get_start_time() + output_actor.get_period()*f_j + output_actor.get_deadline() - (input_actor.get_start_time() + g_i*input_actor.get_period())
                                if latency > self.latency:
                                    self.latency = latency

    def find_sufficient_start_time_vector(self):
        """Finds the sufficient start time vector. Invoked when self.method = SUF"""
        levels = self.find_levels()
        level = 1
        for l in levels:
            for a in l:
                actor = self.get_actor(a.get_actor_id())
                actor.set_start_time((level - 1) * self.alpha)
            level = level + 1

    def find_minimum_buffer_sizes(self):
        """Finds the minimum buffer size vector under the CDP model. Invoked when self.method = MIN"""

        # The following is an optimization to speed-up buffer size computation 
        # in case of implicit-deadline tasks
        if self.get_utilization() == self.get_density(): # Implicit-deadline
            optimize = True
        else:
            optimize = False

        for ch in self.channels:
            source = self.get_actor(ch.get_source())
            destination = self.get_actor(ch.get_destination())
            buffer_size_candidates = list()
            buffer_size_candidates.append(1)
            t_max = max(source.get_start_time(), destination.get_start_time())
            init_val = 0
            if source.get_start_time() <= destination.get_start_time():
                init_val = source.prd(ch, source.get_start_time(), destination.get_start_time())
            if init_val > 0:
                buffer_size_candidates.append(init_val)

            if optimize:
                inc_val = Utilities.gcd(source.get_period(), destination.get_period())
            else:
                inc_val = 1

            for k in range(0, self.alpha + 1, inc_val):
                prd = source.prd(ch, source.get_start_time(), t_max + k)
                cns = destination.cns(ch, destination.get_start_time(), t_max + k)
                unconsumed = prd - cns
                if unconsumed > 0:
                    buffer_size_candidates.append(unconsumed)
            min_buf_size = max(buffer_size_candidates)
            ch.set_buffer_size(min_buf_size)

    def find_sufficient_buffer_sizes(self):
        """Finds the sufficient buffer size vector under the IDP model. Invoked when self.method = IDP_SUF"""
        for ch in self.channels:
            source = self.get_actor(ch.get_source())
            buffer_size = source.prd(ch, source.get_start_time(), source.get_start_time() + self.get_alpha())
            ch.set_buffer_size(max(1, buffer_size))

    def clear(self):
        """Clears the start time and buffer vectors of the graph"""
        for a in self.get_actors():
            a.set_start_time(0)
            a.set_deadline(a.get_period())
        for c in self.get_channels():
            c.set_buffer_size(sys.maxsize)

    def get_execution_vector(self):
        """Return the execution time vector"""
        execution_vector = list()
        for i in self.actors:
            execution_vector.append(i.get_wcet())
        return execution_vector

    def get_execution_vector_str(self):
        """Return the execution time vector as a string"""
        execution_vector = ["["]
        for i in self.actors:
            execution_vector.append("%s_%s = %s, " % (i.get_name(), i.get_function(), i.get_wcet()))
        execution_vector.append("]")
        return ''.join(execution_vector)

    def get_period_vector(self):
        """Return the period vector"""
        period_vector = list()
        for i in self.actors:
            period_vector.append(i.get_period())
        return period_vector

    def get_period_vector_str(self):
        """Return the period vector as a string"""
        period_vector = ["["]
        for i in self.actors:
            period_vector.append("%s_%s = %s, " % (i.get_name(), i.get_function(), i.get_period()))
        period_vector.append("]")
        return ''.join(period_vector)

    def get_buffer_size_vector(self):
        """Return the buffer size vector"""
        buffer_vector = list()
        for c in self.channels:
            buffer_vector.append(c.get_buffer_size())
        return buffer_vector

    def get_buffer_size_vector_str(self):
        """Return the buffer size vector as a string"""
        buffer_vector = ["["]
        for c in self.channels:
            buffer_vector.append("%s = %s, " % (c.get_name(), c.get_buffer_size()))
        buffer_vector += "]"
        return ''.join(buffer_vector)

    def get_start_time_vector(self):
        """Return the start time vector"""
        start_time_vector = list()
        for i in self.actors:
            start_time_vector.append(i.get_start_time())
        return start_time_vector

    def get_start_time_vector_str(self):
        """Return the start time vector as a string"""
        start_time_vector = ["["]
        for i in self.actors:
            start_time_vector.append("%s_%s = %s, " % (i.get_name(), i.get_function(), i.get_start_time()))
        start_time_vector.append("]")
        return ''.join(start_time_vector)

    def get_deadline_vector(self):
        """Return the deadline vector"""
        deadline_vector = list()
        for i in self.actors:
            deadline_vector.append(i.get_deadline())
        return deadline_vector

    def get_deadline_vector_str(self):
        """Return the deadline vector as a string"""
        deadline_vector = ["["]
        for i in self.actors:
            deadline_vector.append("%s_%s = %s, " % (i.get_name(), i.get_function(), i.get_deadline()))
        deadline_vector.append("]")
        return ''.join(deadline_vector)

    def get_repetition_vector(self):
        """Return the repetition vecotr"""
        repetition_vector = list()
        for i in self.actors:
            repetition_vector.append(i.get_repetition())
        return repetition_vector

    def get_repetition_vector_str(self):
        """Return the repetition vecotr as a string"""
        repetition_vector = ["["]
        for i in self.actors:
            repetition_vector.append("%s_%s = %s, " % (i.get_name(), i.get_function(), i.get_repetition()))
        repetition_vector.append("]")
        return ''.join(repetition_vector)

    def find_minimum_repetition_vector(self):
        """Finds the basic repetition vector"""
        q_vec = []
        for i in self.actors:
            q_vec.append(i.get_repetition())
        gcd = Utilities.gcdv(q_vec)
        for i in self.actors:
            r = i.get_repetition()/gcd
            i.set_repetition(r)

    def get_major_cycle(self):
        return self.major_cycle

    def get_minor_cycle(self):
        return self.minor_cycle

    def get_utilization(self):
        """Returns the total utilization"""
        utilization = 0.0
        for i in self.actors:
            utilization = utilization + float(i.get_wcet())/float(i.get_period())
        return utilization

    def get_density(self):
        """Return the total density"""
        density = 0.0
        for i in self.actors:
            density = density + float(i.get_wcet())/min(float(i.get_period()), float(i.get_deadline()))
        return density

    def get_max_utilization(self):
        """Returns the maximum utilization factor"""
        max_utilization = 0.0
        for i in self.actors:
            utilization = float(i.get_wcet())/float(i.get_period())
            if (utilization > max_utilization):
                max_utilization = utilization
        return max_utilization

    def copy_actors(self):
        """Copies the actors list into a new one"""
        new_actors = list()
        for i in self.actors:
            a = ActorModel(self, i.actor_id, i.name, i.function, i.wcet)
            a.inports = list(i.inports)
            a.outports = list(i.outports)
            a.set_repetition(i.get_repetition())
            new_actors.append(a)
        return new_actors

    @staticmethod
    def load_graph_from_file(filename, factor):
        """Loads an ACSDF graph from an gph file"""
        name = os.path.basename(filename).split(".")[0]
        fp = open(filename, 'r')
        cp = CSDFParser(name, fp)
        g = cp.parse_graph()
        g.num_of_levels = len(g.find_levels())
        if verbose:
            sys.stdout.write("Computing the minimum period vector...\n")
        g.find_minimum_repetition_vector()
        g.set_factor(factor)
        g.find_minimum_period_vector(g.get_factor())

        return g

    def calc_start_time_buffer_vector(self, method, deadline_value):
        """Computes the start time and buffer size vector according to the given method"""
        assert method == self.SUF or method == self.MIN
        assert (deadline_value >= 0.0 and deadline_value <= 1.0) 

        if method == self.MIN:
            if verbose:
                sys.stdout.write("Computing the earliest start time vector [CDP - Minimal]...\n")
            self.find_earliest_start_time_vector(deadline_value)
            if verbose:
                sys.stdout.write("Computing the minimum buffer sizes [CDP - Minimal]...\n")
            self.find_minimum_buffer_sizes()
        if method == self.SUF:
            if verbose:
                sys.stdout.write("Computing the earliest start time vector [IDP - Sufficient]...\n")
            self.find_sufficient_start_time_vector()
            if verbose:
                sys.stdout.write("Computing the minimum buffer sizes [IDP - Sufficient]...\n")
            self.find_sufficient_buffer_sizes()

        self.compute_cyclic_executive_parameters()

    @staticmethod
    def print_graph_dot(g, output_dir):
        graph = os.path.join(output_dir, g.get_graph_name() + ".dot")
        fp = open(graph, 'w')
        fp.write("digraph %s{\n" % (g.get_graph_name()))
        fp.write("rankdir=TB;\n")

        levels = g.find_levels()
        for s in levels:
            for a in s:
                actor = g.get_actor(a.get_actor_id())
                for outport in actor.get_outports():
                    ch = g.get_channel(outport)
                    successor = g.get_actor(ch.get_destination())
                    fp.write("%s_%s_%s[style=filled,color=\"lightblue\"];\n" % (g.get_graph_name(), actor.get_name(), actor.get_function()))
                    fp.write("%s_%s_%s[style=filled,color=\"lightblue\"];\n" % (g.get_graph_name(), successor.get_name(), successor.get_function()))
                    fp.write("%s_%s_%s -> %s_%s_%s;\n" % (g.get_graph_name(), actor.get_name(), actor.get_function(), g.get_graph_name(), successor.get_name(), successor.get_function()))
        fp.write('label=\"%s\";\n' % (g.get_graph_name()))
        fp.write("}\n")
        fp.flush()
        fp.close()

    def __repr__(self):
        s = []
        for i in self.actors:
            s.append("%s\n" % (repr(i)))
        for i in self.channels:
            s.append("%s\n" % (repr(i)))
        return ''.join(s)

def main():

    parser = argparse.ArgumentParser(description=__description__, epilog="Author: %s" % __author__)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--minimum", nargs=1, type=float, default=1.0, metavar="factor", help="Compute the minimum start time and buffer size using the given deadline factor. The factor value must be in the range [0,1]. The deadline is computed as follows: D_i = C_i + factor*(T_i - C_i), where D_i is the deadline of task f_i, T_i is the period of f_i, and C_i is the WCET of f_i")
    group.add_argument("--sufficient", action="store_true", help="Compute the sufficient start time and buffer size")
    parser.add_argument("graph", help="The path to the CSDF graph")
    args = parser.parse_args()

    if not os.path.exists(args.graph):
        sys.stderr.write("Error! the provided path does not exist!\n")
        sys.exit(-1)

    if args.sufficient:
        method = ACSDFModel.SUF
        deadline_factor = 0 # Unused in the sufficient case
    else:
        method = ACSDFModel.MIN
        deadline_factor = args.minimum[0]
        if deadline_factor < 0.0 or deadline_factor > 1.0:
            sys.stderr.write("Invalid value for deadline factor! Must be value between 0.0 and 1.0\n")
            sys.exit(-1)

    sys.stdout.write(" - Processing application " + args.graph + "...\n")
    g = ACSDFModel.load_graph_from_file(args.graph, 1)
    g.calc_start_time_buffer_vector(method, deadline_factor)

    sys.stdout.write(" -- Number of actors = %s\n" % (g.get_num_of_actors()))
    sys.stdout.write(" -- Repetition vector = %s\n" % (g.get_repetition_vector_str()))
    sys.stdout.write(" -- Worst-case execution time vector = %s\n" % (g.get_execution_vector_str()))
    sys.stdout.write(" -- Minimum period vector = %s\n" % (g.get_period_vector_str()))
    sys.stdout.write(" -- Deadline vector = %s\n" % (g.get_deadline_vector_str()))
    sys.stdout.write(" -- Start time vector = %s\n" % (g.get_start_time_vector_str()))
    sys.stdout.write(" -- Buffer size vector = %s\n" % (g.get_buffer_size_vector_str()))
    sys.stdout.write(" -- Total Utilization  = %s\n" % (g.get_utilization()))
    sys.stdout.write(" -- Total Density = %s\n" % (g.get_density()))
    sys.stdout.write(" -- Graph maximum latency = %s\n" % (g.get_latency()))

    #ACSDFModel.print_graph_dot(g,os.getcwd())

if __name__ == "__main__":
    main()

