#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Parses a Cyclo-Static Dataflow graph in the .gph format and returns an ACSDFModel object'

import os
import sys
import fractions
import argparse

import Utilities    

class CSDFParser:
    """Parses a CSDF .gph graph and returns an ACSDFModel object"""

    def __init__(self, name, input_file):
        self.input_file = input_file
        self.graph_name = name
        self.graph_type = "csdf"
        self.graph = None

    @staticmethod
    def dump_sdf3_file(sdf3, g):
        sdf3.flush()
        t = 0
        tabs = Utilities.tabs
        # Preamble
        sdf3.write("""<?xml version="1.0" encoding="UTF-8"?>\n""")
        sdf3.write("""<sdf3 type="csdf" version="1.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.es.ele.tue.nl/sdf3/xsd/sdf3-csdf.xsd">\n""")
        t += 1
        sdf3.write("""%s<applicationGraph name=\"%s\">\n""" % (tabs(t), g.get_graph_name()))
        t += 1
        sdf3.write("""%s<csdf name=\"%s\" type=\"%s\">\n""" % (tabs(t), g.get_graph_name(), g.get_graph_name()))
        t += 1
        # Actors
        for node in g.get_actors():
            sdf3.write("\n%s<actor name=\"%s\" type=\"Node\">\n" % (tabs(t), node.get_name()))
            t += 1
            for key, val in node.get_raw_inports().items():
                sdf3.write("%s<port name=\"p_%s\" type=\"in\" rate=\"" % (tabs(t), key))
                for i in range(0,len(val)):
                    if i < (len(val) - 1):
                        sdf3.write("%s," % (val[i]))
                    else:
                        sdf3.write("%s" % (val[i]))
                sdf3.write("\"/>\n")
            for key, val in node.get_raw_outports().items():
                sdf3.write("%s<port name=\"p_%s\" type=\"out\" rate=\"" % (tabs(t), key))
                for i in range(0,len(val)):
                    if i < (len(val) - 1):
                        sdf3.write("%s," % (val[i]))
                    else:
                        sdf3.write("%s" % (val[i]))
                sdf3.write("\"/>\n")
            sdf3.write("""%s<port name="_p1" type="out" rate="1"/>\n%s<port name="_p2" type="in" rate="1"/>\n""" % (tabs(t), tabs(t)))
            t -= 1
            sdf3.write("%s</actor>\n" % (tabs(t)))
        sdf3.write("\n")
        # Channels
        i = 1
        for e in g.get_channels():
            sdf3.write("%s<channel name=\"%s\" srcActor=\"%s\" srcPort=\"p_%s" % (tabs(t), e.get_name(), g.get_actor(e.get_source()).get_name(), e.get_raw_outport()))
            sdf3.write("\" dstActor=\"%s\" dstPort=\"p_%s" % (g.get_actor(e.get_destination()).get_name(), e.get_raw_inport()))
            sdf3.write("\"/>\n")
            i = i + 1
        # Back edges
        i = 1
        for n in g.get_actors():
            sdf3.write("%s<channel name=\"_ch%s\" srcActor=\"%s\" srcPort=\"_p1\" dstActor=\"%s\" dstPort=\"_p2\" initialTokens='1'/>\n" % (tabs(t), i, n.get_name(), n.get_name()))
            i = i + 1
        t -= 1
        sdf3.write("%s</csdf>\n" % (tabs(t)))
        sdf3.write("%s<csdfProperties>\n" % (tabs(t)))
        # Execution times
        t += 1
        for node in g.get_actors():
            sdf3.write("%s<actorProperties actor=\"%s\">\n" % (tabs(t), node.get_name()))
            t += 1
            sdf3.write("%s<processor type=\"p1\" default=\"true\">\n" % (tabs(t)))
            t += 1
            sdf3.write("%s<executionTime time=\"" % (tabs(t)))
            for i in range(0, len(node.get_wcet_sequence())):
                if i < (len(node.get_wcet_sequence())-1):
                    sdf3.write("%s," % (node.get_wcet_sequence()[i]))
                else:
                    sdf3.write("%s" % node.get_wcet_sequence()[i])
            sdf3.write("\"/>\n")
            t -= 1
            sdf3.write("%s</processor>\n" % (tabs(t)))
            t -= 1
            sdf3.write("%s</actorProperties>\n" % (tabs(t)))
        # Closeing part
        t -= 1
        sdf3.write("%s</csdfProperties>\n" % (tabs(t)))
        t -= 1
        sdf3.write("%s</applicationGraph>\n" % (tabs(t)))
        sdf3.write("</sdf3>\n")
        sdf3.flush()

    def read_field(self, field_name):
        oline = self.input_file.readline()
        # Ignore comments
        while oline.startswith('#'):
            oline = self.input_file.readline()
        line = oline.lstrip('\t')
        s = line.split(":")
        if s[0] != field_name:
            # Restore to the previous position
            self.input_file.seek(self.input_file.tell() - len(oline))
            raise Exception("Unable to read field: " % (field_name))
        return s[1].rstrip('\n')

    def read_node(self):
        self.read_field("node")
        actor_id = int(self.read_field("id"))
        actor_name = self.read_field("name")
        function = self.read_field("function")
        seq_length = int(self.read_field("length"))
        wcet_seq = list(map(int, self.read_field("wcet").split(" ")))
        wcet = max(wcet_seq)
        self.graph.add_actor(actor_id, actor_name, function, wcet)
        self.graph.get_actor(actor_id).set_wcet_sequence(wcet_seq)

        num_of_ports = int(self.read_field("port_number"))
        in_ports = dict()
        out_ports = dict()
        for i in range(0,num_of_ports):
            self.read_field("port")
            port_type = self.read_field("type")
            port_id = int(self.read_field("id"))
            p_rate = list(map(int, self.read_field("rate").split()))
            assert len(p_rate) == seq_length
            if port_type == "out":
                out_ports[port_id] = p_rate
            elif port_type == "in":
                in_ports[port_id] = p_rate
            else:
                sys.stderr.write("Invalid port type!")
                raise Exception("Invalid port type!")
        self.graph.get_actor(actor_id).set_raw_inports(in_ports)
        self.graph.get_actor(actor_id).set_raw_outports(out_ports)

    def read_edge(self):
        self.read_field("edge")
        edge_id = int(self.read_field("id"))
        edge_name = self.read_field("name")
        src = self.read_field("src")
        src_s = src.split(" ")
        src_actor = int(src_s[0])
        src_port = int(src_s[1])
        dst = self.read_field("dst")
        dst_s = dst.split(" ")
        dst_actor = int(dst_s[0])
        dst_port = int(dst_s[1])
        self.graph.add_channel(edge_id, edge_name, src_actor, dst_actor, self.graph.get_actor(src_actor).get_raw_outports()[src_port], self.graph.get_actor(dst_actor).get_raw_inports()[dst_port])
        self.graph.get_channel(edge_id).set_raw_inport(dst_port)
        self.graph.get_channel(edge_id).set_raw_outport(src_port)

    def compute_repetition_vector(self):
        """Compute the repetition vector according to the linear time algorithm
        given in page 48 of the book 'Software Synthesis from Dataflow Graphs'"""
        g = self.graph
        self.reps = dict({i:fractions.Fraction(0,1) for i in g.get_actors()})
        a_prime = g.get_actors()[0]
        self.set_reps(a_prime, fractions.Fraction(1,1))
        x_l = []
        for key, val in self.reps.items():
            x_l.append(val.denominator)
        x = fractions.Fraction(int(Utilities.lcmv(x_l)), 1)
        for a in g.get_actors():
            self.reps[a] *= x
        for e in g.get_channels():
            if self.reps[g.get_actor(e.get_source())] * sum(e.get_production_sequence()) \
                != self.reps[g.get_actor(e.get_destination())] * sum(e.get_consumption_sequence()):
                sys.stderr.write("Error! Inconsistent graph\n")
                sys.exit(-1)
        for a in g.get_actors():
            a.set_repetition(self.reps[a].numerator * a.get_P())

    def set_reps(self, a, n):
        g = self.graph
        self.reps[a] = n
        for e in g.get_channels():
            if e.get_source() == a.get_actor_id():
                if self.reps[g.get_actor(e.get_destination())] == fractions.Fraction(0,1):
                    self.set_reps(g.get_actor(e.get_destination()), fractions.Fraction(n *\
                        sum(e.get_production_sequence()), sum(e.get_consumption_sequence())))
        for e in g.get_channels():
            if e.get_destination() == a.get_actor_id():
                if self.reps[g.get_actor(e.get_source())] == fractions.Fraction(0,1):
                    self.set_reps(g.get_actor(e.get_source()), fractions.Fraction(n *\
                        sum(e.get_consumption_sequence()), sum(e.get_production_sequence())))

    def parse_graph(self):
        from ACSDFModel import ACSDFModel
        self.graph = ACSDFModel()

        try:
            self.graph_name = self.read_field("name")
            self.graph_type = self.read_field("type")
        except Exception:
            pass # do nothing. This is an old .gph file with no name and type

        if self.graph_type != "csdf":
            raise Exception("The graph type is not csdf")

        self.graph.set_graph_name(self.graph_name)

        total_nodes = int(self.read_field("node_number"))
        for i in range(0,total_nodes):
            self.read_node()
        total_edges = int(self.read_field("edge_number"))
        for i in range(0, total_edges):
            self.read_edge()
        self.compute_repetition_vector()
        return self.graph


if __name__ == '__main__':

    name = "myApplication"

    parser = argparse.ArgumentParser(description="Reads a CSDF graph in the .gph format from stdin, and dumps the result into stdout", epilog="Author: %s" % __author__)
    parser.add_argument("-s", action="store_true", help="Generate the graph in SDF^3 XML format")
    args = parser.parse_args()

    cp = CSDFParser(name, sys.stdin)
    g = cp.parse_graph()

    if g == None:
        sys.stderr.write("Error encountered while processing the graph!\n")
        sys.exit(-1)

    if not args.s:
        print(g)
    else:
        CSDFParser.dump_sdf3_file(sys.stdout, g)

    sys.exit(0)

