#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama and Jiali Teddy Zhai'
__description__ = 'Generates the task-set representation of a set of graphs and allocates it onto the minimum required number of processors'

import sys
import os
import math
import argparse

from ACSDFModel import ACSDFModel
from Mapping import Mapping
from EspamHandler import EspamHandler
from TextHandler import TextHandler
from DotHandler import DotHandler

import Utilities
import PlatformParameters

# Change verbose to True to print detailed information while running
verbose=False

class PlatformGenerator:

    supported_mappings = {
        "PEDF-FF" : "Partitioned EDF + first-fit allocation",
        "PEDF-FFD" : "Partitioned EDF + first-fit Decreasing allocation",
        "PEDF-WF" : "Partitioned EDF + worst-fit allocation",
        #"PEDF-WFD" : "Partitioned EDF + worst-fit Decreasing allocation",
        #"PEDF-BFD" : "Partitioned EDF + best-fit Decreasing allocation",
        "RM-FF" : "Partitioned Fixed-Priority + Rate-Monotonic priority + first-fit allocation using exact utilization bound",
        "RM-LCE" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using exhaustive search and exact utilization bound",
        "RM-LCR" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using random search and exact utilization bound",
        "RM-LCG" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using genetic algorithm and exact utilization bound"
        }

    schedulers_map = {
        'PEDF-FF' : Mapping.find_PEDF_FF,
        'PEDF-FFD' : Mapping.find_PEDF_FFD,
        'PEDF-WF' : Mapping.find_PEDF_WF,
        #'PEDF-WFD' : Mapping.find_PEDF_WFD,
        #'PEDF-BFD' : Mapping.find_PEDF_BFD,
        'RM-FF' : Mapping.find_RM_FF,
        'RM-LCE' : Mapping.find_RM_LCE,
        'RM-LCR' : Mapping.find_RM_LCR,
        'RM-LCG' : Mapping.find_RM_LCG,
        }

    def __init__(self):
        self.partitions = []
        self.scheduler_type = "PEDF-FF"
        self.map = None
        self.graphs = []
        self.tasks = []
        self.total_utilization = 0.0
        self.max_utilization = 0.0
        self.total_density = 0.0
        self.max_density = 0.0
        self.num_of_apps = len(self.graphs)
        self.total_utilization = 0.0
        self.beta = 0.0
        self.min_num_of_processors = -1
    
    # process graphs parameters
    def process_graphs(self, apps, factors, token_sizes, ignorestartbuf, printinfo=False):
        # compute periods with given the scaling factor
        graphs = list()
        for i in range(0, len(apps)):
            g = ACSDFModel.load_graph_from_file(apps[i], 1)
            g.set_factor(factors[i])
            g.set_max_token_size(token_sizes[i])
            if printinfo:
                sys.stdout.write("\n")
                sys.stdout.write(" - Processing application %s...\n" % (os.path.basename(apps[i])))
                sys.stdout.write(" -- Number of actors = %s\n" % (g.get_num_of_actors()))
                sys.stdout.write(" -- Repetition vector = %s\n" % (g.get_repetition_vector_str()))
                sys.stdout.write(" -- Worst-case execution time vector = %s\n" % (g.get_execution_vector_str()))
                sys.stdout.write(" -- Minimum period vector = %s\n" % (g.get_period_vector_str()))
                sys.stdout.write(" -- Deadline vector = %s\n" % (g.get_deadline_vector_str()))
            if verbose:
                sys.stdout.write("\n")
                sys.stdout.write(" -- Least common multiple of the repetitions (Q) = %s\n" % (g.lcm)) 
                sys.stdout.write(" -- max(WCET_i * q_i) (eta) = %s\n" % (g.max_q_C)) 
                sys.stdout.write(" -- Iteration period = %s\n" % (g.alpha))
                if g.get_matched_io_type():
                    sys.stdout.write(" -- %s is a matched I/O rates graph\n" % (g.get_graph_name()))
                else:
                    sys.stdout.write(" -- %s is a mis-matched I/O rates graph\n" % (g.get_graph_name()))
                if g.get_minor_cycle() != 0:
                    sys.stdout.write(" -- Cyclic executive major cycle = " + str(g.get_major_cycle()) + ", Cyclic Executive Minor Cycle = " + str(g.get_minor_cycle()) + "\n")
                else:
                    sys.stdout.write(" -- Unable to find a minor cycle with no preemptions!\n")            
                sys.stdout.write(" -- Hyperperiod = %s\n" % (Utilities.lcmv(g.get_period_vector())))

            g.update_parameters(PlatformParameters.AXI_CB, PlatformParameters.T_RD_AXI_CB, PlatformParameters.T_WR_AXI_CB)
            if not ignorestartbuf:
                g.calc_start_time_buffer_vector(ACSDFModel.MIN, 1)
            graphs.append(g)

        if printinfo:
            sys.stdout.write("\n")
        
        assert len(graphs) > 0
        self.graphs = graphs
        self.num_of_apps = len(self.graphs)
        num_of_tasks = 0
        for g in self.graphs:
            for a in g.get_actors():
                a.set_task_id(num_of_tasks)
                self.total_utilization = self.total_utilization + a.get_utilization()
                self.total_density = self.total_density + a.get_density()
                self.tasks.append(a)
                num_of_tasks += 1
                if a.get_utilization() > self.max_utilization:
                    self.max_utilization = a.get_utilization()
                if a.get_density() > self.max_density:
                    self.max_density = a.get_density()
        self.beta = int(math.floor(1.0/self.max_utilization))
        if self.total_utilization != self.total_density:
            self.min_num_of_processors = int(math.ceil(self.total_density))
        else:
            self.min_num_of_processors = int(math.ceil(self.total_utilization))
    
    def get_min_num_of_processors(self):
        return self.min_num_of_processors

    def get_num_of_tasks(self):
        return len(self.tasks)

    def get_tasks(self):
        return self.tasks

    def get_scheduler_type(self):
        return self.scheduler_type
    
    def set_scheduler_type(self, scheduler_type):
        self.scheduler_type = scheduler_type
        self.map = self.schedulers_map[scheduler_type]
    
    def get_graphs(self):
        return self.graphs

    def get_num_of_graphs(self):
        return len(self.graphs)

    def get_task_by_name(self, task_name):
        for t in self.get_tasks():
            if t.get_name() == task_name:
                return t
        return None

    def get_partitions(self):
        return self.partitions

    def set_partitions(self, partitions):
        self.partitions = partitions

    def update_graphs(self, mapping, optimized=True):
        amapping = [set() for i in range(len(mapping))]
        j = 0
        for s in mapping:
            for t in s:
                if type(t) == int:
                    task = self.get_tasks()[t]
                else:
                    task = t
                task.set_affinity(j)
                amapping[j].add(task)
            j += 1
        channels = self.get_mapped_channels(amapping)
        gamma =[[0 for i in range(len(mapping))] for j in range(len(mapping))]
        for c in channels:
            gamma[c[0]][c[1]] = 1

        # Set the contention
        for a in self.get_tasks():
            if optimized:
                contention = sum(gamma[a.get_affinity()])
            else:
                contention = len(mapping)
            a.set_contention(contention)
        for g in self.get_graphs():
            g.update_parameters(PlatformParameters.AXI_CB, PlatformParameters.T_RD_AXI_CB, PlatformParameters.T_WR_AXI_CB)

    def get_mapped_channels(self,partitions):
        channels = []
        chan_cnt = 0
        for a in self.get_tasks():
            actor = a
            for outport in actor.get_outports():
                ch = actor.get_graph().get_channel(outport)
                successor = actor.get_graph().get_actor(ch.get_destination())
                channels.append((actor.get_affinity(), successor.get_affinity()))
                chan_cnt += 1
        return channels


def make_graphs(apps, factors, token_sizes, ignorestartbuf = False, printinfo=False):
    assert len(factors) == len(apps)    
    pg = PlatformGenerator()
    pg.process_graphs(apps, factors, token_sizes, ignorestartbuf, printinfo)
    return pg
    

def make_platform(output_dir, pg, scheduler_type, ignorestartbuf=False, printinfo=False):
    # scheduler type
    pg.set_scheduler_type(scheduler_type)
        
    # Mapping phase
    partitions = pg.map(pg)

    # In case of exhaustive search, the results is actually a list of the best mappings
    if pg.get_scheduler_type().startswith("RM-LCE"):
        partitions = partitions[-1]

    pg.update_graphs(partitions)
    pg.set_partitions(partitions)
    if verbose:
        sys.stdout.write(" - Contention of the mapping = %s\n" % (Mapping.contention(partitions,pg)))
        sys.stdout.write(" - Mapping is %s" % (partitions))
        sys.stdout.write("\n\n")
    # Update start times and buffer sizes
    for g in pg.get_graphs():
        if not ignorestartbuf:
            g.calc_start_time_buffer_vector(ACSDFModel.MIN, 1)
        if printinfo:
            sys.stdout.write(" - Dumping %s information...\n" % (g.get_graph_name()))
            sys.stdout.write(" -- Period vector = %s\n" % (g.get_period_vector_str()))
            sys.stdout.write(" -- Start time vector = %s\n" % (g.get_start_time_vector_str()))
            sys.stdout.write(" -- Buffer size vector = %s\n" % (g.get_buffer_size_vector_str()))
            sys.stdout.write(" -- Sink actors periods:")
            g_levels = g.find_levels()
            for aa in g_levels[-1]:
                ac = g.get_actor(aa.get_actor_id())
                sys.stdout.write(" %s (%s) : %s\n" % (ac.get_full_name(), ac.get_function(), ac.get_period()))
            sys.stdout.write(" -- Total Utilization  = %s\n" % (g.get_utilization()))
            sys.stdout.write(" -- Graph maximum latency = %s\n" % (g.get_latency()))
            sys.stdout.write("\n")

    if printinfo:
        sys.stdout.write(" - Minimum number of processors needed to schedule the applications by:\n")
        sys.stdout.write(" -- Optimal scheduling algorithms (M_OPT) = %s\n" % (pg.get_min_num_of_processors()))
        sys.stdout.write(" -- %s (M_%s) = %s\n" % (PlatformGenerator.supported_mappings[pg.get_scheduler_type()], pg.get_scheduler_type(), len(partitions)))
        sys.stdout.write("\n")

    # Call ESPAM handler
    eh = EspamHandler(pg, 'ML605', output_dir, partitions)
    eh.generate_platform_file()
    eh.generate_mapping_file()

    # Call text handler
    th = TextHandler(pg, output_dir, partitions)
    th.generate_mapping_file()

    dh = DotHandler(pg, output_dir, partitions)
    dh.generate_mapping_file()

    return pg

def main():

    parser = argparse.ArgumentParser(description=__description__, epilog="Authors: %s" % __author__)
    required = parser.add_argument_group("required arguments")
    required.add_argument("-o", required=True, nargs=1, type=str, metavar="directory", help="The output directory where the platform and mapping specifications will be generated")
    parser.add_argument("-c", action="store_true", help="Output the time values in cycles instead of OS clock ticks")
    required.add_argument("-m", required=True, nargs=1, metavar="algorithm", type=str, help="The scheduling and mapping algorithm")
    required.add_argument("-a", required=True, action="append", nargs=3, metavar=("scaling-factor", "max-token-size", "graph"),\
                        help="Specifies an application that will be processed. scaling-factor is an integer >=1 that is used to scale the minimum period. max-token-size is the maximum token size (in 32-bit words). graph is the path to the .gph file containing the CSDF model of the application. Multiple applications can be specified by passing -a multiple times. For example: -a 1 10 app1.gph -a 2 1 app2.gph -a 10 1 app3.gph")
    
    args = parser.parse_args()

    if not os.path.isdir(args.o[0]):
        sys.stderr.write("Error! The specified output directory does not exist!\n")
        sys.exit(-1)
    
    output_dir = args.o[0]

    if not args.c:
        PlatformParameters.TIME_IN_CYCLES = False

    if args.m[0] not in PlatformGenerator.supported_mappings:
        sys.stderr.write("ERROR! Invalid mapping algorithm specified\n")
        sys.exit(-1)    

    mapping_algo = args.m[0]

    factors = []
    token_sizes = []
    apps = []

    for a in args.a:
        factors.append(int(a[0]))
        token_sizes.append(int(a[1]))
        apps.append(a[2])
    
    pg = make_graphs(apps, factors, token_sizes)
    make_platform(output_dir, pg, mapping_algo, printinfo=True)

if __name__ == "__main__":
    main()

